/*
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Martin Husemann
 *	and Wolfgang Solfrank.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <sys/cdefs.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/errno.h>

#define	DEV_BSIZE	512
#define DOSBOOTBLOCKSIZE DEV_BSIZE

typedef	u_int32_t	cl_t;	/* type holding a cluster number */

struct bootblock {
	u_int	BytesPerSec;		/* bytes per sector */
	u_int	SecPerClust;		/* sectors per cluster */
	u_int	ResSectors;		/* number of reserved sectors */
	u_int	FATs;			/* number of FATs */
	u_int	RootDirEnts;		/* number of root directory entries */
	u_int	Media;			/* media descriptor */
	u_int	FATsmall;		/* number of sectors per FAT */
	u_int	SecPerTrack;		/* sectors per track */
	u_int	Heads;			/* number of heads */
	u_int32_t Sectors;		/* total number of sectors */
	u_int32_t HiddenSecs;		/* # of hidden sectors */
	u_int32_t HugeSectors;		/* # of sectors if bpbSectors == 0 */
	u_int	FSInfo;			/* FSInfo sector */
	u_int	Backup;			/* Backup of Bootblocks */
	cl_t	RootCl;			/* Start of Root Directory */
	cl_t	FSFree;			/* Number of free clusters acc. FSInfo */
	cl_t	FSNext;			/* Next free cluster acc. FSInfo */

	/* and some more calculated values */
	u_int	flags;			/* some flags: */
#define	FAT32		1		/* this is a FAT32 filesystem */
					/*
					 * Maybe, we should separate out
					 * various parts of FAT32?	XXX
					 */
	int	ValidFat;		/* valid fat if FAT32 non-mirrored */
	cl_t	ClustMask;		/* mask for entries in FAT */
	cl_t	NumClusters;		/* # of entries in a FAT */
	u_int32_t NumSectors;		/* how many sectors are there */
	u_int32_t FATsecs;		/* how many sectors are in FAT */
	u_int32_t NumFatEntries;	/* how many entries really are there */
	u_int	ClusterOffset;		/* at what sector would sector 0 start */
	u_int	ClusterSize;		/* Cluster size in bytes */

	/* Now some statistics: */
	u_int	NumFiles;		/* # of plain files */
	u_int	NumFree;		/* # of free clusters */
	u_int	NumBad;			/* # of bad clusters */
};

struct fatEntry {
	cl_t	next;			/* pointer to next cluster */
	cl_t	head;			/* pointer to start of chain */
	u_int32_t length;		/* number of clusters on chain */
	int	flags;			/* see below */
};

#define	CLUST_FREE	0		/* 0 means cluster is free */
#define	CLUST_FIRST	2		/* 2 is the minimum valid cluster number */
#define	CLUST_RSRVD	0xfffffff6	/* start of reserved clusters */
#define	CLUST_BAD	0xfffffff7	/* a cluster with a defect */
#define	CLUST_EOFS	0xfffffff8	/* start of EOF indicators */
#define	CLUST_EOF	0xffffffff	/* standard value for last cluster */

/*
 * Masks for cluster values
 */
#define	CLUST12_MASK	0xfff
#define	CLUST16_MASK	0xffff
#define	CLUST32_MASK	0xfffffff

#define	FAT_USED	1		/* This fat chain is used in a file */

#define	DOSLONGNAMELEN	256		/* long name maximal length */
#define LRFIRST		0x40		/* first long name record */
#define	LRNOMASK	0x1f		/* mask to extract long record
					 * sequence number */

/*
 * Architecture independent description of a directory entry
 */
struct dosDirEntry {
	struct dosDirEntry
		*parent,		/* previous tree level */
		*next,			/* next brother */
		*child;			/* if this is a directory */
	char name[8+1+3+1];		/* alias name first part */
	char lname[DOSLONGNAMELEN];	/* real name */
	uint flags;			/* attributes */
	cl_t head;			/* cluster no */
	u_int32_t size;			/* filesize in bytes */
	uint fsckflags;			/* flags during fsck */
};

/* Flags in fsckflags: */
#define	DIREMPTY	1
#define	DIREMPWARN	2

static int do_ddfs __P((const char *, u_long));
static int checkclnum __P((struct bootblock *, int, cl_t, cl_t *));
static int readfat __P((int, struct bootblock *, int, struct fatEntry **));
static int checkfat __P((struct bootblock *, struct fatEntry *));
static int writefat __P((int, struct bootblock *, struct fatEntry *));
static int allocate_contigblks __P((struct bootblock *, struct fatEntry *, u_long, cl_t *));
static int connect __P((int, struct bootblock *, struct fatEntry *, cl_t, cl_t));
static int readboot __P((int, struct bootblock *));
static int writefsinfo __P((int, struct bootblock *));

static char *myprog;

int
main(argc, argv)
	int argc;
	char **argv;
{
	int ret, ch;
	u_long size = 0;

	myprog = argv[0];
	while ((ch = getopt(argc, argv, "vs:")) != EOF)
	{
		switch (ch)
		{
		case 'v':
			break;
		case 's':
			size = (u_long) atol(optarg);
			break;
		default:
			exit(0);
		}
	}

	argc -= optind;
	argv += optind;
	
	size = size * (1024 * 1024 / DEV_BSIZE);
	ret = do_ddfs(argv[0], size);
	return ret;
}

static int
do_ddfs(fname, nblks)
	const char *fname;
	u_long nblks;
{
	int dosfs;
	struct bootblock boot;
	struct fatEntry *fat = NULL;
	int fatno;
	cl_t startcl, rootdircl;

	dosfs = open(fname, O_RDWR, 0);
	if (dosfs < 0)
	{
		perror(myprog);
		exit(1);
	} 

	if (readboot(dosfs, &boot) != 0)
	{
		close(dosfs);
		return 8;
	}

	fatno = 0;
	if (readfat(dosfs, &boot, fatno, &fat) != 0)
	{	
		close(dosfs);
		return 8;
	}

	if (checkfat(&boot, fat) != 0)
	{
bad:
		free(fat);
		close(dosfs);
		return 8;
	}

	if (allocate_contigblks(&boot, fat, nblks, &startcl) != 0)
		goto bad;

	if (writefat(dosfs, &boot, fat) != 0)
		goto bad;

	if (boot.flags & FAT32)
		rootdircl = boot.RootCl;
	else
		rootdircl = 0;

	if (connect(dosfs, &boot, fat, startcl, rootdircl) != 0)
		goto bad;

	if (writefsinfo(dosfs, &boot) != 0)
		goto bad;	

	free(fat);
	close(dosfs);
	return 0;
}

static int
checkclnum(boot, fat, cl, next)
	struct bootblock *boot;
	int fat;
	cl_t cl;
	cl_t *next;
{
	if (*next >= (CLUST_RSRVD&boot->ClustMask))
		*next |= ~boot->ClustMask;

	if (*next == CLUST_FREE) {
		boot->NumFree++;
		return 0;
	}

	if (*next == CLUST_BAD) {
		boot->NumBad++;
		return 0;
	}

	if (*next < CLUST_FIRST || 
	    (*next >= boot->NumClusters && *next < CLUST_EOFS)) {
		return EINVAL;
	}
	return 0;
}

/*
 * Read a FAT and decode it into internal format
 */
static int
readfat(fs, boot, no, fp)
	int fs;
	struct bootblock *boot;
	int no;
	struct fatEntry **fp;
{
	struct fatEntry *fat;
	u_char *buffer, *p;
	cl_t cl;
	off_t off;
	int ret = 0;

	boot->NumFree = boot->NumBad = 0;
	fat = malloc(sizeof(struct fatEntry) * boot->NumClusters);
	buffer = malloc(boot->FATsecs * boot->BytesPerSec);
	if (fat == NULL || buffer == NULL) {
		perror("No space for FAT");
		if (fat)
			free(fat);
		return EINVAL;
	}

	memset(fat, 0, sizeof(struct fatEntry) * boot->NumClusters);

	off = boot->ResSectors + no * boot->FATsecs;
	off *= boot->BytesPerSec;

	if (lseek(fs, off, SEEK_SET) != off) {
		perror("Unable to read FAT");
		free(buffer);
		free(fat);
		return EINVAL;
	}

	if (read(fs, buffer, boot->FATsecs * boot->BytesPerSec)
	    != boot->FATsecs * boot->BytesPerSec) {
		perror("Unable to read FAT");
		free(buffer);
		free(fat);
		return EINVAL;
	}

#if	0
	if (buffer[0] != boot->Media || 
#else
	if (
#endif
		buffer[1] != 0xff || buffer[2] != 0xff ||
		(boot->ClustMask == CLUST16_MASK && buffer[3] != 0xff) ||
		(boot->ClustMask == CLUST32_MASK
		&& ((buffer[3]&0x0f) != 0x0f
		    || buffer[4] != 0xff || buffer[5] != 0xff
		    || buffer[6] != 0xff || (buffer[7]&0x0f) != 0x0f)))
	{
		printf("%s: illegal fat\n");
		return EINVAL;
	}

	switch (boot->ClustMask) {
	case CLUST32_MASK:
		p = buffer + 8;
		break;
	case CLUST16_MASK:
		p = buffer + 4;
		break;
	default:
		p = buffer + 3;
		break;
	}
	for (cl = CLUST_FIRST; cl < boot->NumClusters;) {
		switch (boot->ClustMask) {
		case CLUST32_MASK:
			fat[cl].next = p[0] + (p[1] << 8)
				       + (p[2] << 16) + (p[3] << 24);
			fat[cl].next &= boot->ClustMask;
			ret |= checkclnum(boot, no, cl, &fat[cl].next);
			cl++;
			p += 4;
			break;
		case CLUST16_MASK:
			fat[cl].next = p[0] + (p[1] << 8);
			ret |= checkclnum(boot, no, cl, &fat[cl].next);
			cl++;
			p += 2;
			break;
		default:
			fat[cl].next = (p[0] + (p[1] << 8)) & 0x0fff;
			ret |= checkclnum(boot, no, cl, &fat[cl].next);
			cl++;
			if (cl >= boot->NumClusters)
				break;
			fat[cl].next = ((p[1] >> 4) + (p[2] << 4)) & 0x0fff;
			ret |= checkclnum(boot, no, cl, &fat[cl].next);
			cl++;
			p += 3;
			break;
		}
	}

	free(buffer);
	*fp = fat;
	return ret;
}

static int
checkfat(boot, fat)
	struct bootblock *boot;
	struct fatEntry *fat;
{
	cl_t head, p, h, n;
	u_int len;
	int ret = 0;

	/*
	 * pass 1: figure out the cluster chains.
	 */
	for (head = CLUST_FIRST; head < boot->NumClusters; head++)
	{
		if (fat[head].head != 0		
		    || fat[head].next == CLUST_FREE
		    || fat[head].next == CLUST_BAD)
			continue;	

		for (len = 0, p = head;
		     p >= CLUST_FIRST && p < boot->NumClusters;
		     p = fat[p].next) {
			fat[p].head = head;
			len++;
		}

		/* the head record gets the length */
		fat[head].length = fat[head].next == CLUST_FREE ? 0 : len;
	}

	/*
	 * pass 2: check for crosslinked chains (we couldn't do this in pass 1 because
	 * we didn't know the real start of the chain then - would have treated partial
	 * chains as interlinked with their main chain)
	 */
	for (head = CLUST_FIRST; head < boot->NumClusters; head++) {
		/* find next untravelled chain */
		if (fat[head].head != head)
			continue;

		/* follow the chain to its end (hopefully) */
		for (p = head;
		     (n = fat[p].next) >= CLUST_FIRST && n < boot->NumClusters;
		     p = n)
			if (fat[n].head != head)
				break;
		if (n >= CLUST_EOFS)
			continue;

		printf("file systems inconsistent\n");
		return EINVAL;
	}
	return 0;
}

/*
 * Write out FATs encoding them from the internal format
 */
static int
writefat(fs, boot, fat)
	int fs;
	struct bootblock *boot;
	struct fatEntry *fat;
{
	u_char *buffer, *p;
	cl_t cl;
	int i;
	u_int32_t fatsz;
	off_t off;
	int ret = 0;

	buffer = malloc(fatsz = boot->FATsecs * boot->BytesPerSec);
	if (buffer == NULL) {
		perror("No space for FAT");
		return EINVAL;
	}

	off = boot->ResSectors;  
	off *= boot->BytesPerSec;
	if (lseek(fs, off, SEEK_SET) != off)
	{
		goto badread;
	}

	if (read(fs, buffer, fatsz) != fatsz)
	{
badread:
		perror("Unable to read FAT");
		free(buffer);
		free(fat);
		return EINVAL;
	}

	boot->NumFree = 0;
	p = buffer;
#if	0
	*p++ = (u_char)boot->Media;
#else	
	p ++;
#endif
	*p++ = 0xff;
	*p++ = 0xff;
	switch (boot->ClustMask) {
	case CLUST16_MASK:
		*p++ = 0xff;
		break;
	case CLUST32_MASK:
		*p++ = 0x0f;
		*p++ = 0xff;
		*p++ = 0xff;
		*p++ = 0xff;
		*p++ = 0x0f;
		break;
	}
	for (cl = CLUST_FIRST; cl < boot->NumClusters; cl++) {
		switch (boot->ClustMask) {
		case CLUST32_MASK:
			if (fat[cl].next == CLUST_FREE)
				boot->NumFree++;
			*p++ = (u_char)fat[cl].next;
			*p++ = (u_char)(fat[cl].next >> 8);
			*p++ = (u_char)(fat[cl].next >> 16);
			*p &= 0xf0;
			*p++ |= (fat[cl].next >> 24)&0x0f;
			break;
		case CLUST16_MASK:
			if (fat[cl].next == CLUST_FREE)
				boot->NumFree++;
			*p++ = (u_char)fat[cl].next;
			*p++ = (u_char)(fat[cl].next >> 8);
			break;
		default:
			if (fat[cl].next == CLUST_FREE)
				boot->NumFree++;
			if (cl + 1 < boot->NumClusters
			    && fat[cl + 1].next == CLUST_FREE)
				boot->NumFree++;
			*p++ = (u_char)fat[cl].next;
			*p++ = (u_char)((fat[cl].next >> 8) & 0xf)
			       |(u_char)(fat[cl+1].next << 4);
			*p++ = (u_char)(fat[++cl].next >> 4);
			break;
		}
	}
	for (i = 0; i < boot->FATs; i++) {
		off = boot->ResSectors + i * boot->FATsecs;
		off *= boot->BytesPerSec;
		if (lseek(fs, off, SEEK_SET) != off
		    || write(fs, buffer, fatsz) != fatsz) {
			perror("Unable to write FAT");
			ret = EINVAL; /* Return immediately?		XXX */
		}
	}
	free(buffer);
	return ret;
}

static int
allocate_contigblks(boot, fat, nblks, startclp)
	struct bootblock *boot;
	struct fatEntry *fat;
	u_long nblks;
	cl_t *startclp;
{
	u_long nent, maxcontig, startcl, reqcl;
	cl_t ohead, head;

	nent = maxcontig = 0;
	startcl = CLUST_FIRST;
	for (ohead = head = CLUST_FIRST; head < boot->NumClusters; head++)
	{
		if (fat[head].head != 0	|| fat[head].next != CLUST_FREE)
		{
			if (nent >= maxcontig)
			{
				maxcontig = nent;
				startcl = ohead;
			}
			ohead = head + 1;
			nent = 0;
			continue;
		}
		nent ++;
	}

	if (nent >= maxcontig)
	{
		maxcontig = nent;
		startcl = ohead;
	}

	if (maxcontig < 2 ||
	    startcl + maxcontig > boot->NumClusters)
	{
		printf("%s: no contig clusters\n", myprog);
		return ENOSPC;
	}

	if (nblks == 0)
		reqcl = maxcontig;
	else
		reqcl = nblks / ((boot->BytesPerSec * boot->SecPerClust) / DEV_BSIZE);

	if (maxcontig < reqcl)
	{
		printf("%s: no spaces left %ld < %ld\n",
			myprog, maxcontig, reqcl);
		return ENOSPC;
	}

	if (maxcontig > reqcl)
	{
		startcl = startcl + (maxcontig - reqcl);
		maxcontig = reqcl;
	}

	for (nent = startcl; nent < startcl + maxcontig; nent ++)
	{
		if (fat[nent].next != CLUST_FREE)
		{
			printf("%s: missing contig sectors\n", myprog);
			return EINVAL;
		}
		fat[nent].head  = startcl;
		fat[nent].next = nent + 1;
	}

	fat[nent - 1].next = CLUST_EOFS;
	fat[startcl].length = maxcontig;
	*startclp = startcl;
	return 0;
}

/*
 * Try to reconnect a FAT chain into dir
 */
#define	SLOT_EMPTY	0x00		/* slot has never been used */
#define	SLOT_E5		0x05		/* the real value is 0xe5 */
#define	SLOT_DELETED	0xe5		/* file in this slot deleted */

static u_char *lfbuf;
static cl_t lfcl;
static off_t lfoff;

static int
connect(dosfs, boot, fat, head, dirhead)
	int dosfs;
	struct bootblock *boot;
	struct fatEntry *fat;
	cl_t head, dirhead;
{
	struct dosDirEntry d;
	u_char *p;

	if (lfbuf == NULL)
	{
		lfbuf = malloc(boot->ClusterSize);
		if (!lfbuf) {
			perror("No space for buffer");
			return EINVAL;
		}
	}

	p = NULL;

	while (1) {
		if (p != NULL)
		{
			for (; p < lfbuf + boot->ClusterSize; p += 32)
			{
				if (*p == SLOT_EMPTY || *p == SLOT_DELETED)
					break;
			}
		}

		if (p != NULL && p < lfbuf + boot->ClusterSize)
			break;

		if (dirhead == 0)	/* FAT 12/16 root */
		{
			lfcl = 0;
			lfoff = boot->ResSectors + boot->FATs * boot->FATsecs;
			lfoff *= boot->BytesPerSec;
		}
		else
		{
			lfcl = (p != NULL) ? fat[lfcl].next : dirhead;
			if (lfcl < CLUST_FIRST || lfcl >= boot->NumClusters)
			{
				printf("%s: no spaces left\n", myprog);
				return ENOSPC;
			}
			lfoff = lfcl * boot->ClusterSize
			    + boot->ClusterOffset * boot->BytesPerSec;
		}

		if (lseek(dosfs, lfoff, SEEK_SET) != lfoff
		    || read(dosfs, lfbuf, boot->ClusterSize) != boot->ClusterSize) {
			perror("could not read LOST.DIR");
			return EINVAL;
		}
		p = lfbuf;
	}

	boot->NumFiles++;
	/* Ensure uniqueness of entry here!				XXX */
	memset(&d, 0, sizeof d);
	(void)snprintf(d.name, sizeof(d.name), "%s", "NETBSDFS");
	d.flags = 0;
	d.head = head;
	d.size = fat[head].length * boot->ClusterSize;

	memset(p, 0, 32);
	memset(p, ' ', 11);
	memcpy(p, d.name, strlen(d.name));
	p[11] = 0x05;	/* ATTR_READONLY | ATTR_SYSTEM */
	p[26] = (u_char)d.head;
	p[27] = (u_char)(d.head >> 8);
	if (boot->ClustMask == CLUST32_MASK) {
		p[20] = (u_char)(d.head >> 16);
		p[21] = (u_char)(d.head >> 24);
	}
	p[28] = (u_char)d.size;
	p[29] = (u_char)(d.size >> 8);
	p[30] = (u_char)(d.size >> 16);
	p[31] = (u_char)(d.size >> 24);
	fat[head].flags |= FAT_USED;
	if (lseek(dosfs, lfoff, SEEK_SET) != lfoff
	    || write(dosfs, lfbuf, boot->ClusterSize) != boot->ClusterSize) {
		perror("could not write directory");
		return EINVAL;
	}
	return 0;
}

/************************************************************
 * Read boot area
 ************************************************************/
static int
readboot(dosfs, boot)
	int dosfs;
	struct bootblock *boot;
{
	u_char block[DOSBOOTBLOCKSIZE];
	u_char fsinfo[2 * DOSBOOTBLOCKSIZE];
	u_char backup[DOSBOOTBLOCKSIZE];
	int ret = 0;
#ifndef	ORIGINAL_CODE
	u_int sec_shift;
#endif	/* PC98 */
	
	if (read(dosfs, block, sizeof block) < sizeof block) {
		perror("could not read boot block");
		return EINVAL;
	}

	if (block[510] != 0x55 || block[511] != 0xaa) {
		printf("Invalid signature in boot block: %02x%02x", block[511], block[510]);
		return EINVAL;
	}

	memset(boot, 0, sizeof *boot);
	boot->ValidFat = -1;

	/* decode bios parameter block */
	boot->BytesPerSec = block[11] + (block[12] << 8);
#ifndef	ORIGINAL_CODE
	sec_shift = boot->BytesPerSec / 512;
	boot->BytesPerSec = 512;
#endif	/* PC-98 */
	boot->SecPerClust = block[13];
#ifndef	ORIGINAL_CODE
	boot->SecPerClust *= sec_shift;
#endif	/* PC98 */
	boot->ResSectors = block[14] + (block[15] << 8);
#ifndef	ORIGINAL_CODE
	boot->ResSectors *= sec_shift;
#endif	/* PC98 */
	boot->FATs = block[16];
	boot->RootDirEnts = block[17] + (block[18] << 8);
	boot->Sectors = block[19] + (block[20] << 8);
#ifndef	ORIGINAL_CODE
	boot->Sectors *= sec_shift;
#endif	/* PC98 */
	boot->Media = block[21];
	boot->FATsmall = block[22] + (block[23] << 8);
	boot->SecPerTrack = block[24] + (block[25] << 8);
#ifndef	ORIGINAL_CODE
	boot->FATsmall *= sec_shift;
	boot->SecPerTrack *= sec_shift;
#endif	/* PC98 */
	boot->Heads = block[26] + (block[27] << 8);
	boot->HiddenSecs = block[28] + (block[29] << 8) + (block[30] << 16) + (block[31] << 24);
	boot->HugeSectors = block[32] + (block[33] << 8) + (block[34] << 16) + (block[35] << 24);
#ifndef	ORIGINAL_CODE
	boot->HiddenSecs *= sec_shift;
	boot->HugeSectors *= sec_shift;
#endif	/* PC98 */

	boot->FATsecs = boot->FATsmall;

	if (!boot->RootDirEnts)
		boot->flags |= FAT32;
	if (boot->flags & FAT32) {
		boot->FATsecs = block[36] + (block[37] << 8)
				+ (block[38] << 16) + (block[39] << 24);
		if (block[40] & 0x80)
			boot->ValidFat = block[40] & 0x0f;

		/* check version number: */
		if (block[42] || block[43]) {
			/* Correct?				XXX */
			printf("Unknown filesystem version: %x.%x",
			       block[43], block[42]);
			return EINVAL;
		}
		boot->RootCl = block[44] + (block[45] << 8)
			       + (block[46] << 16) + (block[47] << 24);
		boot->FSInfo = block[48] + (block[49] << 8);
		boot->Backup = block[50] + (block[51] << 8);

		if (lseek(dosfs, boot->FSInfo * boot->BytesPerSec, SEEK_SET)
		    != boot->FSInfo * boot->BytesPerSec
		    || read(dosfs, fsinfo, sizeof fsinfo)
		    != sizeof fsinfo) {
			perror("could not read fsinfo block");
			return EINVAL;
		}
		if (memcmp(fsinfo, "RRaA", 4)
		    || memcmp(fsinfo + 0x1e4, "rrAa", 4)
		    || fsinfo[0x1fc]
		    || fsinfo[0x1fd]
		    || fsinfo[0x1fe] != 0x55
		    || fsinfo[0x1ff] != 0xaa
		    || fsinfo[0x3fc]
		    || fsinfo[0x3fd]
		    || fsinfo[0x3fe] != 0x55
		    || fsinfo[0x3ff] != 0xaa)
		{
			return EINVAL;
		}
		if (boot->FSInfo) {
			boot->FSFree = fsinfo[0x1e8] + (fsinfo[0x1e9] << 8)
				       + (fsinfo[0x1ea] << 16)
				       + (fsinfo[0x1eb] << 24);
			boot->FSNext = fsinfo[0x1ec] + (fsinfo[0x1ed] << 8)
				       + (fsinfo[0x1ee] << 16)
				       + (fsinfo[0x1ef] << 24);
		}

		if (lseek(dosfs, boot->Backup * boot->BytesPerSec, SEEK_SET)
		    != boot->Backup * boot->BytesPerSec
		    || read(dosfs, backup, sizeof backup) != sizeof  backup) {
			perror("could not read backup bootblock");
			return EINVAL;
		}
		if (memcmp(block, backup, DOSBOOTBLOCKSIZE)) {
			/* Correct?					XXX */
			printf("backup doesn't compare to primary bootblock");
			return EINVAL;
		}
		/* Check backup FSInfo?					XXX */
	}

	boot->ClusterOffset = (boot->RootDirEnts * 32 + boot->BytesPerSec - 1)
	    / boot->BytesPerSec
	    + boot->ResSectors
	    + boot->FATs * boot->FATsecs
	    - CLUST_FIRST * boot->SecPerClust;

	if (boot->BytesPerSec % DOSBOOTBLOCKSIZE != 0) {
		printf("Invalid sector size: %u", boot->BytesPerSec);
		return EINVAL;
	}
	if (boot->SecPerClust == 0) {
		printf("Invalid cluster size: %u", boot->SecPerClust);
		return EINVAL;
	}
	if (boot->Sectors) {
		boot->HugeSectors = 0;
		boot->NumSectors = boot->Sectors;
	} else
		boot->NumSectors = boot->HugeSectors;
	boot->NumClusters = (boot->NumSectors - boot->ClusterOffset) / boot->SecPerClust;

	if (boot->flags&FAT32)
		boot->ClustMask = CLUST32_MASK;
	else if (boot->NumClusters < (CLUST_RSRVD&CLUST12_MASK))
		boot->ClustMask = CLUST12_MASK;
	else if (boot->NumClusters < (CLUST_RSRVD&CLUST16_MASK))
		boot->ClustMask = CLUST16_MASK;
	else {
		printf("Filesystem too big (%u clusters) for non-FAT32 partition",
		       boot->NumClusters);
		return EINVAL;
	}

	switch (boot->ClustMask) {
	case CLUST32_MASK:
		boot->NumFatEntries = (boot->FATsecs * boot->BytesPerSec) / 4;
		break;
	case CLUST16_MASK:
		boot->NumFatEntries = (boot->FATsecs * boot->BytesPerSec) / 2;
		break;
	default:
		boot->NumFatEntries = (boot->FATsecs * boot->BytesPerSec * 2) / 3;
		break;
	}

	if (boot->NumFatEntries < boot->NumClusters) {
		printf("FAT size too small, %u entries won't fit into %u sectors\n",
		       boot->NumClusters, boot->FATsecs);
		return EINVAL;
	}
	boot->ClusterSize = boot->BytesPerSec * boot->SecPerClust;

	boot->NumFiles = 1;
	boot->NumFree = 0;

	return ret;
}

static int
writefsinfo(dosfs, boot)
	int dosfs;
	struct bootblock *boot;
{
	u_char block[DOSBOOTBLOCKSIZE];
	u_char fsinfo[2 * DOSBOOTBLOCKSIZE];

	if (lseek(dosfs, (off_t) 0, SEEK_SET) != (off_t) 0 ||
	    read(dosfs, block, sizeof block) < sizeof block) 
	{
		perror("could not read boot block");
		return EINVAL;
	}

#if	0
	block[17] = boot->RootDirEnts & 0xff;
	block[18] = (boot->RootDirEnts >> 8) & 0xff;

	if (lseek(dosfs, (off_t) 0, SEEK_SET) != (off_t) 0 ||
	    write(dosfs, block, sizeof block) < sizeof block) 
	{
		perror("could not write boot block");
		return EINVAL;
	}
#endif

	if (boot->RootDirEnts == 0)
	{
		if (lseek(dosfs, boot->FSInfo * boot->BytesPerSec, SEEK_SET)
		    != boot->FSInfo * boot->BytesPerSec
		    || read(dosfs, fsinfo, sizeof fsinfo) != sizeof fsinfo) {
			perror("could not read fsinfo block");
			return EINVAL;
		}

		fsinfo[0x1e8] = (u_char)boot->FSFree;
		fsinfo[0x1e9] = (u_char)(boot->FSFree >> 8);
		fsinfo[0x1ea] = (u_char)(boot->FSFree >> 16);
		fsinfo[0x1eb] = (u_char)(boot->FSFree >> 24);
		fsinfo[0x1ec] = (u_char)boot->FSNext;
		fsinfo[0x1ed] = (u_char)(boot->FSNext >> 8);
		fsinfo[0x1ee] = (u_char)(boot->FSNext >> 16);
		fsinfo[0x1ef] = (u_char)(boot->FSNext >> 24);
		if (lseek(dosfs, boot->FSInfo * boot->BytesPerSec, SEEK_SET)
		    != boot->FSInfo * boot->BytesPerSec
		    || write(dosfs, fsinfo, sizeof fsinfo)
		    != sizeof fsinfo) {
			perror("Unable to write FSInfo");
			return EINVAL;
		}
	}
	/*
	 * Technically, we should return FSBOOTMOD here.
	 *
	 * However, since Win95 OSR2 (the first M$ OS that has
	 * support for FAT32) doesn't maintain the FSINFO block
	 * correctly, it has to be fixed pretty often.
	 *
	 * Therefor, we handle the FSINFO block only informally,
	 * fixing it if neccessary, but otherwise ignoring the
	 * fact that it was incorrect.
	 */
	return 0;
}
