/*	$NetBSD$	*/
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 *
 * Polling driver for National Semiconductor DS8390/WD83C690 based
 * ethernet adapters.
 *
 * Copyright (c) 1998 Matthias Drochner.  All rights reserved.
 *
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * Copyright (C) 1993, David Greenman.  This software may be used, modified,
 * copied, distributed, and sold, in both source and binary form provided that
 * the above copyright and these terms are retained.  Under no circumstances is
 * the author responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with its use.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <machine/pio.h>
#include <machine/cpu.h>

#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>

#include <lib/libsa/stand.h>
#include <lib/libsa/net.h>

#include <dev/ic/dp8390reg.h>
#include <dev/ic/ne2000reg.h>
#include "../busio_user.h"
#include "dp8390piosubr.h"

#include "etherdrv.h"

static struct dp8390pio_softc *dp8390pio_soft_gc;

static void dp8390_init __P((struct dp8390pio_softc *));
static int dp8390pio_start __P((struct dp8390pio_softc *, char *, int));
static int dp8390pio_pkt_read __P((struct dp8390pio_softc *, char *, int));
static void dp8390_data_read __P((struct dp8390pio_softc *, int, char *, int));
static void	dp8390pio_writemem __P((bus_space_tag_t, bus_space_handle_t,
	    bus_space_tag_t, bus_space_handle_t, u_int8_t *, int, size_t, int));
static void	dp8390pio_readmem __P((bus_space_tag_t, bus_space_handle_t,
	    bus_space_tag_t, bus_space_handle_t, int, u_int8_t *, size_t, int));

static void
dp8390_init(sc)
	struct dp8390pio_softc *sc;
{
	bus_space_tag_t nict = sc->sc_nict;
	bus_space_handle_t nich = sc->sc_nich;
	int i;

	/*
	 * Initialize the NIC in the exact order outlined in the NS manual.
	 * This init procedure is "mandatory"...don't change what or when
	 * things happen.
	 */

	/* Set interface for page 0, remote DMA complete, stopped. */
	bus_space_write_1(nict, nich, ED_P0_CR,
			  sc->cr_proto | ED_CR_PAGE_0 | ED_CR_STP);

	if (sc->dcr_reg & ED_DCR_LS) {
		bus_space_write_1(nict, nich, ED_P0_DCR, sc->dcr_reg);
	} else {
		/*
		 * Set FIFO threshold to 8, No auto-init Remote DMA, byte
		 * order=80x86, byte-wide DMA xfers,
		 */
		bus_space_write_1(nict, nich, ED_P0_DCR, ED_DCR_FT1 | ED_DCR_LS);
	}

	/* Clear remote byte count registers. */
	bus_space_write_1(nict, nich, ED_P0_RBCR0, 0);
	bus_space_write_1(nict, nich, ED_P0_RBCR1, 0);

	/* Tell RCR to do nothing for now. */
	bus_space_write_1(nict, nich, ED_P0_RCR, ED_RCR_MON);

	/* Place NIC in internal loopback mode. */
	bus_space_write_1(nict, nich, ED_P0_TCR, ED_TCR_LB0);

	/* Initialize receive buffer ring. */
	bus_space_write_1(nict, nich, ED_P0_BNRY, sc->rec_page_start);
	bus_space_write_1(nict, nich, ED_P0_PSTART, sc->rec_page_start);
	bus_space_write_1(nict, nich, ED_P0_PSTOP, sc->rec_page_stop);

	/*
	 * Clear all interrupts.  A '1' in each bit position clears the
	 * corresponding flag.
	 */
	bus_space_write_1(nict, nich, ED_P0_ISR, 0xff);

	/*
	 * Disable all interrupts.
	 */
	bus_space_write_1(nict, nich, ED_P0_IMR, 0);

	/* Program command register for page 1. */
	bus_space_write_1(nict, nich, ED_P0_CR,
			  sc->cr_proto | ED_CR_PAGE_1 | ED_CR_STP);

	/* Copy out our station address. */
	for (i = 0; i < 6; ++i)
		bus_space_write_1(nict, nich, ED_P1_PAR0 + i,
				  sc->eth_myaddr[i]);

	/*
	 * Set current page pointer to one page after the boundary pointer, as
	 * recommended in the National manual.
	 */
	sc->next_packet = sc->rec_page_start + 1;
	bus_space_write_1(nict, nich, ED_P1_CURR, sc->next_packet);

	/* Program command register for page 0. */
	bus_space_write_1(nict, nich, ED_P1_CR,
			  sc->cr_proto | ED_CR_PAGE_0 | ED_CR_STP);

	/* directed and broadcast */
	bus_space_write_1(nict, nich, ED_P0_RCR, ED_RCR_AB);

	/* Take interface out of loopback. */
	bus_space_write_1(nict, nich, ED_P0_TCR, 0);

	/* Fire up the interface. */
	bus_space_write_1(nict, nich, ED_P0_CR,
			  sc->cr_proto | ED_CR_PAGE_0 | ED_CR_STA);
}

int
dp8390_config(sc, membase, memsize)
	struct dp8390pio_softc *sc;
	bus_size_t membase;
	bus_size_t memsize;
{
	bus_space_tag_t nict = sc->sc_nict;
	bus_space_handle_t nich = sc->sc_nich;
	bus_space_tag_t asict = sc->sc_asict;
	bus_space_handle_t asich = sc->sc_asich;
	u_int8_t tmp, romdata[16];
	int i;

	dp8390pio_soft_gc = sc;
	sc->snd_page_start = membase >> ED_PAGE_SHIFT;
	sc->rec_page_start = sc->snd_page_start + ED_TXBUF_SIZE;
	sc->rec_page_stop = sc->snd_page_start + (memsize >> ED_PAGE_SHIFT);

	/* Reset the board. */
	tmp = bus_space_read_1(asict, asich, NE2000_ASIC_RESET);
	delay(10000);
	bus_space_write_1(asict, asich, NE2000_ASIC_RESET, tmp);
	delay(5000);

	bus_space_write_1(nict, nich, ED_P0_CR,
	    ED_CR_RD2 | ED_CR_PAGE_0 | ED_CR_STP);
	delay(5000);

	tmp = bus_space_read_1(nict, nich, ED_P0_CR);
	if ((tmp & (ED_CR_RD2 | ED_CR_TXP | ED_CR_STA | ED_CR_STP)) !=
	    (ED_CR_RD2 | ED_CR_STP))
		return EINVAL;

	(void) bus_space_read_1(nict, nich, ED_P0_ISR);
	bus_space_write_1(nict, nich,
	    ED_P0_CR, ED_CR_RD2 | ED_CR_PAGE_0 | ED_CR_STA);

	for (i = 0; i < 100; i++)
	{
		if ((bus_space_read_1(nict, nich, ED_P0_ISR) & ED_ISR_RST) ==
		    ED_ISR_RST)
		{
			/* Ack the reset bit. */
			bus_space_write_1(nict, nich, ED_P0_ISR, ED_ISR_RST);
			break;
		}
		delay(100);
	}

	bus_space_write_1(nict, nich, ED_P0_RCR, ED_RCR_MON);
	bus_space_write_1(nict, nich, ED_P0_DCR, sc->dcr_reg);
	bus_space_write_1(nict, nich, ED_P0_PSTART, membase >> ED_PAGE_SHIFT);
	bus_space_write_1(nict, nich, ED_P0_PSTOP,
		          (membase + memsize) >> ED_PAGE_SHIFT);

  {
	static u_int8_t test_pattern[32] = "THIS is A memory TEST pattern";
	bus_addr_t bottom = membase + memsize - ED_PAGE_SIZE;
	u_int8_t test_buffer[32];

	dp8390pio_writemem(nict, nich, asict, asich, test_pattern, membase,
			   sizeof(test_pattern), sc->sc_useword);
	dp8390pio_readmem(nict, nich, asict, asich, membase, test_buffer,
		    	  sizeof(test_buffer), sc->sc_useword);
	if (bcmp(test_pattern, test_buffer, sizeof(test_pattern)) != 0)
		return EINVAL;

	dp8390pio_writemem(nict, nich, asict, asich, test_pattern,
			   bottom, sizeof(test_pattern), sc->sc_useword);
	dp8390pio_readmem(nict, nich, asict, asich, bottom, test_buffer,
			  sizeof(test_buffer), sc->sc_useword);
	if (bcmp(test_pattern, test_buffer, sizeof(test_pattern)) != 0)
		return EINVAL;
    }

	bus_space_write_1(nict, nich, ED_P0_ISR, 0xff);
	if (sc->eth_myaddr[0] == 0xff)
	{
		dp8390pio_readmem(nict, nich, asict, asich, sc->sc_romstart,
				  romdata, sizeof(romdata), sc->sc_useword);
		for (i = 0; i < 6; i++)
		{
			sc->eth_myaddr[i] = romdata[i * (sc->sc_useword ? 2 : 1)];
		}
	}

        printf("Ether address %s\n", ether_sprintf(sc->eth_myaddr));

	dp8390_init(sc);
	return(0);
}

void
dp8390_stop(sc)
	struct dp8390pio_softc *sc;
{
	bus_space_tag_t nict = sc->sc_nict;
	bus_space_handle_t nich = sc->sc_nich;
	int n = 5000;

	/* Stop everything on the interface, and select page 0 registers. */
	bus_space_write_1(nict, nich, ED_P0_CR,
			  sc->cr_proto | ED_CR_PAGE_0 | ED_CR_STP);

	/*
	 * Wait for interface to enter stopped state, but limit # of checks to
	 * 'n' (about 5ms).  It shouldn't even take 5us on modern DS8390's, but
	 * just in case it's an old one.
	 */
	while ((((bus_space_read_1(nict, nich, ED_P0_ISR))& ED_ISR_RST) == 0) && --n);
}

int
EtherSend(pkt, len)
	char *pkt;
	int len;
{
	struct dp8390pio_softc *sc = dp8390pio_soft_gc;

	return dp8390pio_start(sc, pkt, len);
}

int
EtherReceive(pkt, maxlen)
	char *pkt;
	int maxlen;
{
	struct dp8390pio_softc *sc = dp8390pio_soft_gc;
	
	return dp8390pio_pkt_read(sc, pkt, maxlen);
}

static int
dp8390pio_start(sc, pkt, len)
	struct dp8390pio_softc *sc;
	char *pkt;
	int len;
{
	bus_space_tag_t nict = sc->sc_nict;
	bus_space_handle_t nich = sc->sc_nich;
	bus_space_tag_t asict = sc->sc_asict;
	bus_space_handle_t asich = sc->sc_asich;

	dp8390pio_writemem(nict, nich, asict, asich, pkt, 
			   sc->snd_page_start << ED_PAGE_SHIFT,
			   len, sc->sc_useword);

	/* Set TX buffer start page. */
	bus_space_write_1(nict, nich, ED_P0_TPSR, sc->snd_page_start);

	/* Set TX length. */
	bus_space_write_1(nict, nich, ED_P0_TBCR0, len < 60 ? 60 : len);
	bus_space_write_1(nict, nich, ED_P0_TBCR1, len >> 8);

	/* Set page 0, remote DMA complete, transmit packet, and *start*. */
	bus_space_write_1(nict, nich, ED_P0_CR,
			  sc->cr_proto | ED_CR_PAGE_0 | ED_CR_TXP | ED_CR_STA);

	return(len);
}

void
dp8390_data_read(sc, buf, dest, len)
	struct dp8390pio_softc *sc;
	int buf;
	char *dest;
	int len;
{
	bus_space_tag_t nict = sc->sc_nict;
	bus_space_handle_t nich = sc->sc_nich;
	bus_space_tag_t asict = sc->sc_asict;
	bus_space_handle_t asich = sc->sc_asich;
	u_int lastp = sc->rec_page_stop << ED_PAGE_SHIFT;
	int tmp_amount;

	if (buf + len > lastp)
	{
		tmp_amount = lastp - buf;

		/* Copy amount up to end of NIC memory. */
		dp8390pio_readmem(nict, nich, asict, asich, buf,
				  dest, tmp_amount, sc->sc_useword);

		len -= tmp_amount;
		buf = (sc->rec_page_start << ED_PAGE_SHIFT);
		dest += tmp_amount;
	}

	dp8390pio_readmem(nict, nich, asict, asich, buf,
			  dest, len, sc->sc_useword);
}

int
dp8390pio_pkt_read(sc, pkt, maxlen)
	struct dp8390pio_softc *sc;
	char *pkt;
	int maxlen;
{
	bus_space_tag_t nict = sc->sc_nict;
	bus_space_handle_t nich = sc->sc_nich;
	bus_space_tag_t asict = sc->sc_asict;
	bus_space_handle_t asich = sc->sc_asich;
	struct dp8390_ring packet_hdr;
	int packet_ptr;
	u_int len;
	u_char boundary, current;

	if (!(bus_space_read_1(nict, nich, ED_P0_RSR) & ED_RSR_PRX))
		return(0); /* XXX error handling */

	/* Set NIC to page 1 registers to get 'current' pointer. */
	bus_space_write_1(nict, nich, ED_P0_CR,
			  sc->cr_proto | ED_CR_PAGE_1 | ED_CR_STA);

	/*
	 * 'sc->next_packet' is the logical beginning of the ring-buffer - i.e.
	 * it points to where new data has been buffered.  The 'CURR' (current)
	 * register points to the logical end of the ring-buffer - i.e. it
	 * points to where additional new data will be added.  We loop here
	 * until the logical beginning equals the logical end (or in other
	 * words, until the ring-buffer is empty).
	 */
	current = bus_space_read_1(nict, nich, ED_P1_CURR);

	/* Set NIC to page 0 registers to update boundary register. */
	bus_space_write_1(nict, nich, ED_P1_CR,
			  sc->cr_proto | ED_CR_PAGE_0 | ED_CR_STA);

	if (sc->next_packet == current)
		return(0);

	/* Get pointer to this buffer's header structure. */
	packet_ptr = (sc->next_packet << ED_PAGE_SHIFT);

	/*
	 * The byte count includes a 4 byte header that was added by
	 * the NIC.
	 */
	dp8390pio_readmem(nict, nich, asict, asich, packet_ptr,
			  (u_int8_t *) &packet_hdr, 4, sc->sc_useword);

	len = packet_hdr.count;
	if (packet_hdr.next_packet < sc->rec_page_start ||
	    packet_hdr.next_packet >= sc->rec_page_stop)
	{
		printf("dp8390_pkt_read: bad pos 0x%x len 0x%x\n",
			(u_int) packet_hdr.next_packet, (u_int) len);
		dp8390_stop(sc);
		dp8390_init(sc);
		return 0;
	}

	len -= sizeof(struct dp8390_ring);
	if (len > ((u_int) maxlen))
		len = maxlen;

	dp8390_data_read(sc, packet_ptr + sizeof(struct dp8390_ring),
			 pkt, len);

	/* Update next packet pointer. */
	sc->next_packet = packet_hdr.next_packet;

	/*
	 * Update NIC boundary pointer - being careful to keep it one
	 * buffer behind (as recommended by NS databook).
	 */
	boundary = sc->next_packet - 1;
	if (boundary < sc->rec_page_start)
		boundary = sc->rec_page_stop - 1;
	bus_space_write_1(nict, nich, ED_P0_BNRY, boundary);
	return(len);
}

static void
dp8390pio_readmem(nict, nich, asict, asich, src, dst, amount, useword)
	bus_space_tag_t nict;
	bus_space_handle_t nich;
	bus_space_tag_t asict;
	bus_space_handle_t asich;
	int src;
	u_int8_t *dst;
	size_t amount;
	int useword;
{

	/* Select page 0 registers. */
	bus_space_write_1(nict, nich, ED_P0_CR,
	    ED_CR_RD2 | ED_CR_PAGE_0 | ED_CR_STA);

	/* Round up to a word. */
	if (amount & 1)
		++amount;

	/* Set up DMA byte count. */
	bus_space_write_1(nict, nich, ED_P0_RBCR0, amount);
	bus_space_write_1(nict, nich, ED_P0_RBCR1, amount >> 8);

	/* Set up source address in NIC mem. */
	bus_space_write_1(nict, nich, ED_P0_RSAR0, src);
	bus_space_write_1(nict, nich, ED_P0_RSAR1, src >> 8);

	bus_space_write_1(nict, nich, ED_P0_CR,
	    ED_CR_RD0 | ED_CR_PAGE_0 | ED_CR_STA);

	if (useword)
		bus_space_read_multi_2(asict, asich, NE2000_ASIC_DATA,
		    (u_int16_t *)dst, amount >> 1);
#ifdef	notyet
	else
		bus_space_read_multi_1(asict, asich, NE2000_ASIC_DATA,
		    dst, amount);
#endif
}

/*
 * Stripped down routine for writing a linear buffer to NIC memory.  Only
 * used in the probe routine to test the memory.  'len' must be even.
 */
static void
dp8390pio_writemem(nict, nich, asict, asich, src, dst, len, useword)
	bus_space_tag_t nict;
	bus_space_handle_t nich;
	bus_space_tag_t asict;
	bus_space_handle_t asich;
	u_int8_t *src;
	int dst;
	size_t len;
	int useword;
{
	int maxwait = 100;	/* about 120us */

	/* Select page 0 registers. */
	bus_space_write_1(nict, nich, ED_P0_CR,
	    ED_CR_RD2 | ED_CR_PAGE_0 | ED_CR_STA);

	/* Reset remote DMA complete flag. */
	bus_space_write_1(nict, nich, ED_P0_ISR, ED_ISR_RDC);

	/* Set up DMA byte count. */
	bus_space_write_1(nict, nich, ED_P0_RBCR0, len);
	bus_space_write_1(nict, nich, ED_P0_RBCR1, len >> 8);

	/* Set up destination address in NIC mem. */
	bus_space_write_1(nict, nich, ED_P0_RSAR0, dst);
	bus_space_write_1(nict, nich, ED_P0_RSAR1, dst >> 8);

	/* Set remote DMA write. */
	bus_space_write_1(nict, nich, ED_P0_CR,
	    ED_CR_RD1 | ED_CR_PAGE_0 | ED_CR_STA);

	if (useword)
		bus_space_write_multi_2(asict, asich, NE2000_ASIC_DATA,
		    (u_int16_t *)src, len >> 1);
#ifdef	notyet
	else
		bus_space_write_multi_1(asict, asich, NE2000_ASIC_DATA,
		    src, len);
#endif

	/*
	 * Wait for remote DMA to complete.  This is necessary because on the
	 * transmit side, data is handled internally by the NIC in bursts, and
	 * we can't start another remote DMA until this one completes.  Not
	 * waiting causes really bad things to happen - like the NIC wedging
	 * the bus.
	 */
	while (((bus_space_read_1(nict, nich, ED_P0_ISR) & ED_ISR_RDC) !=
	    ED_ISR_RDC) && --maxwait);

	if (maxwait == 0)
		printf("dp8390pio_writemem: failed to complete\n");
}
